import typer
from typing_extensions import Annotated

from focus_converter.data_loaders.data_loader import DataFormats, ParquetDataFormat

PROVIDER_OPTION = Annotated[
    str,
    typer.Option(help="Target provider"),
]

EXPORT_INCLUDE_SOURCE_COLUMNS = Annotated[
    bool,
    typer.Option(help="Include source columns", rich_help_panel="Data Export"),
]
EXPORT_PATH_OPTION = Annotated[
    str, typer.Option(help="Target data path", rich_help_panel="Data Export")
]

DATA_PATH = Annotated[
    str, typer.Option(help="Source data path", rich_help_panel="Source Data")
]
DATA_FORMAT_OPTION = Annotated[
    DataFormats, typer.Option(help="Data format", rich_help_panel="Source Data")
]
PARQUET_DATA_FORMAT_OPTION = Annotated[
    ParquetDataFormat,
    typer.Option(help="Parquet data format", rich_help_panel="Source Data"),
]

PLAN_GRAPH_PATH = Annotated[
    str,
    typer.Option(
        help="Path to save plan graph",
        rich_help_panel="Plan Graph",
    ),
]



from typing import List, Tuple

import polars as pl

from focus_converter.configs.base_config import (
    ConversionPlan,
    MissingColumnDType,
    SetColumnDTypesConversionArgs,
)
from focus_converter.conversion_functions.validations import ColumnValidator


class DeferredColumnFunctions:
    """
    # Set of functions that can only be executed once the lazyframe is loaded and the column names available
    # are known.
    """

    def __init__(self):
        # missing_column_plans = []
        self.__missing_column_plans__: List[Tuple[str, ConversionPlan]] = []

        # enforced column dtypes, need to be applied before any other conversion
        # if column is present then a cast operation can be applied, if not then a new column can be added
        # with null values and then cast operation can be applied
        self.__enforced_column_dtypes__: List[ConversionPlan] = []

    @staticmethod
    def convert_focus_data_type_polars_dtype(focus_data_type):
        if focus_data_type == "string":
            return pl.Utf8
        elif focus_data_type == "float":
            return pl.Float64
        elif focus_data_type == "int":
            return pl.Int64
        elif focus_data_type == "datetime":
            return pl.Datetime
        elif focus_data_type == "date":
            return pl.Date
        else:
            raise RuntimeError(f"data_type: {focus_data_type} not implemented")

    def map_missing_column_plan(
        self, plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ):
        self.__missing_column_plans__.append((column_alias, plan))
        column_validator.map_static_default_value_if_not_present(
            plan=plan, column_alias=column_alias
        )

    def map_dtype_plan(self, plan: ConversionPlan, column_validator: ColumnValidator):
        self.__enforced_column_dtypes__.append(plan)
        column_validator.map_dtype_enforced_node(plan=plan)

    def apply_missing_column_plan(self, lf: pl.LazyFrame):
        for column_alias, missing_column_plan in self.__missing_column_plans__:
            if missing_column_plan.column not in lf.columns:
                conversion_arg: MissingColumnDType = MissingColumnDType.model_validate(
                    missing_column_plan.conversion_args
                )

                if conversion_arg.data_type == "string":
                    dtype = pl.Utf8
                elif conversion_arg.data_type == "float":
                    dtype = pl.Float64
                elif conversion_arg.data_type == "int":
                    dtype = pl.Int64
                else:
                    raise RuntimeError(
                        f"data_type: {conversion_arg.data_types} not implemented"
                    )

                lf = lf.with_columns(
                    pl.lit(None).cast(dtype).alias(missing_column_plan.column)
                )
            else:
                lf = lf.with_columns(
                    pl.col(missing_column_plan.column).alias(missing_column_plan.column)
                )
        return lf

    def apply_dtype_plan(self, lf: pl.LazyFrame):
        for plan in self.__enforced_column_dtypes__:
            conversion_args = SetColumnDTypesConversionArgs.model_validate(
                plan.conversion_args
            )
            for column_obj in conversion_args.dtype_args:
                if column_obj.column_name not in lf.columns:
                    lf = lf.with_columns(
                        pl.lit(None)
                        .cast(
                            self.convert_focus_data_type_polars_dtype(column_obj.dtype),
                            strict=False,
                        )
                        .alias(column_obj.column_name)
                    )
                else:
                    # check if the column is i64 for the cast to work else fail, at the point the column
                    # should be used with datetime parser plan.
                    cast_type = self.convert_focus_data_type_polars_dtype(
                        column_obj.dtype
                    )
                    if cast_type in [pl.Datetime, pl.Date]:
                        if lf.schema[column_obj.column_name] in [pl.Datetime, pl.Date]:
                            # ignore if the column is already of type datetime/date
                            pass
                        elif lf.schema[column_obj.column_name] == pl.Utf8:
                            if cast_type == pl.Datetime:
                                lf = lf.with_columns(
                                    pl.col(column_obj.column_name).str.to_datetime()
                                )
                            else:
                                lf = lf.with_columns(
                                    pl.col(column_obj.column_name).str.to_date()
                                )
                        else:
                            # possibly a timestamp column
                            lf = lf.with_columns(
                                pl.col(column_obj.column_name).cast(
                                    self.convert_focus_data_type_polars_dtype(
                                        column_obj.dtype
                                    ),
                                    strict=True,
                                )
                            )
                    else:
                        lf = lf.with_columns(
                            pl.col(column_obj.column_name).cast(
                                self.convert_focus_data_type_polars_dtype(
                                    column_obj.dtype
                                ),
                                strict=False,
                            )
                        )
        return lf



from enum import Enum
from typing import Annotated

import pandas as pd
import typer
import yaml

from focus_converter.configs.base_config import ConversionPlan
from focus_converter.conversion_functions import STATIC_CONVERSION_TYPES
from focus_converter.converter import FocusConverter
from focus_converter.models.focus_column_names import FocusColumnNames

app = typer.Typer(name="FOCUS export conversion rules", add_completion=False)


class ReportFormats(Enum):
    CSV = "csv"
    MARKDOWN = "markdown"


@app.command("export", help="Converts source cost data to FOCUS format")
def export_conversion_rules(
    output_dir: Annotated[
        str,
        typer.Option(help="Write provider csvs to this path."),
    ],
    output_format: Annotated[
        ReportFormats,
        typer.Option(help="Output format"),
    ],
):
    # iterate through providers
    converter = FocusConverter()
    converter.load_provider_conversion_configs()

    for provider, plans in converter.plans.items():
        rows = []
        converter.prepare_horizontal_conversion_plan(provider=provider)

        # count number of times a dimension is transformed
        transform_step_counter = {}

        source_column_dtypes = {}
        for plan in plans:
            if (
                plan.conversion_type
                == STATIC_CONVERSION_TYPES.APPLY_DEFAULT_IF_COLUMN_MISSING
            ):
                source_column_dtypes[plan.column] = plan.conversion_args["data_type"]
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.SET_COLUMN_DTYPES:
                for dtype_arg in plan.conversion_args["dtype_args"]:
                    source_column_dtypes[dtype_arg["column_name"]] = dtype_arg["dtype"]

        plan: ConversionPlan
        for i, plan in enumerate(plans):
            # skip if dtype plan
            if plan.conversion_type in [
                STATIC_CONVERSION_TYPES.SET_COLUMN_DTYPES,
                STATIC_CONVERSION_TYPES.APPLY_DEFAULT_IF_COLUMN_MISSING,
            ]:
                continue

            try:
                transform_step_counter[plan.focus_column] += 1
            except KeyError:
                transform_step_counter[plan.focus_column] = 1

            if plan.conversion_type in [
                STATIC_CONVERSION_TYPES.SQL_CONDITION,
                STATIC_CONVERSION_TYPES.ASSIGN_STATIC_VALUE,
                STATIC_CONVERSION_TYPES.LOOKUP,
                STATIC_CONVERSION_TYPES.ASSIGN_STATIC_VALUE,
                STATIC_CONVERSION_TYPES.MAP_VALUES,
            ]:
                formatted_conversion_args = yaml.dump(plan.conversion_args)
            elif (
                plan.conversion_type
                == STATIC_CONVERSION_TYPES.APPLY_DEFAULT_IF_COLUMN_MISSING
            ):
                formatted_conversion_args = yaml.dump(plan.conversion_args["data_type"])
            else:
                formatted_conversion_args = plan.conversion_args

            rows.append(
                {
                    "FOCUS Dimension": plan.focus_column.value,
                    "Transform Step": transform_step_counter[plan.focus_column],
                    "Source Column": plan.column,
                    "Source Column Type": source_column_dtypes.get(
                        plan.column, "Not Defined"
                    ),
                    "Transform Type": plan.conversion_type.name,
                    "Filters/Process/Etc.": formatted_conversion_args,
                }
            )

        # add missing output columns plan
        for focus_column in FocusColumnNames:
            if (
                focus_column not in transform_step_counter
                and focus_column != FocusColumnNames.PLACE_HOLDER
            ):
                rows.append(
                    {
                        "FOCUS Dimension": focus_column.value,
                        "Transform Step": 0,
                        "Source Column": "Not Defined",
                        "Source Column Type": "Not Defined",
                        "Transform Type": "Not Defined",
                        "Filters/Process/Etc.": "Not Defined",
                    }
                )

        # sort rows by transform step and focus dimension
        rows = sorted(
            rows,
            key=lambda x: (x["Transform Step"] > 0, x["FOCUS Dimension"]),
        )

        df = pd.DataFrame(rows)
        if output_format == ReportFormats.CSV:
            df.to_csv(f"{output_dir}/{provider}.csv", index=False)
        elif output_format == ReportFormats.MARKDOWN:
            df.to_markdown(f"{output_dir}/{provider}.md", index=False)
        else:
            raise ValueError(f"Invalid output format: {output_format}")


if __name__ == "__main__":
    app()



from enum import Enum

import polars as pl

DEFAULT_FOCUS_NAMESPACE = "F"


class FocusColumnNames(Enum):
    """
    Focus column names as described in https://focus.finops.org/#specification
    """

    PLACE_HOLDER = "PlaceHolder"

    CHARGE_PERIOD_START = "ChargePeriodStart"
    CHARGE_PERIOD_END = "ChargePeriodEnd"
    CHARGE_FREQUENCY = "ChargeFrequency"
    CHARGE_SUB_CATEGORY = "ChargeSubcategory"
    CHARGE_DESCRIPTION = "ChargeDescription"

    BILLING_PERIOD_START = "BillingPeriodStart"
    BILLING_PERIOD_END = "BillingPeriodEnd"

    PROVIDER = "Provider"
    PUBLISHER = "Publisher"
    INVOICE_ISSUER = "InvoiceIssuer"

    BILLING_ACCOUNT_ID = "BillingAccountId"
    BILLING_ACCOUNT_NAME = "BillingAccountName"

    BILLED_COST = "BilledCost"
    BILLED_CURRENCY = "BilledCurrency"
    BILLING_CURRENCY = "BillingCurrency"
    EFFECTIVE_COST = "EffectiveCost"

    REGION = "Region"

    CHARGE_TYPE = "ChargeType"

    SERVICE_CATEGORY = "ServiceCategory"
    SERVICE_NAME = "ServiceName"

    SUB_ACCOUNT_NAME = "SubAccountName"
    SUB_ACCOUNT_ID = "SubAccountId"

    AVAILABILITY_ZONE = "AvailabilityZone"

    RESOURCE_NAME = "ResourceName"
    RESOURCE_ID = "ResourceId"
    RESOURCE_TYPE = "ResourceType"

    COMMITMENT_DISCOUNT_NAME = "CommitmentDiscountName"
    COMMITMENT_DISCOUNT_ID = "CommitmentDiscountId"
    COMMITMENT_DISCOUNT_TYPE = "CommitmentDiscountType"
    COMMITMENT_DISCOUNT_CATEGORY = "CommitmentDiscountCategory"

    PRICING_QUANTITY = "PricingQuantity"
    PRICING_UNIT = "PricingUnit"
    PRICING_CATEGORY = "PricingCategory"

    SKU_ID = "SkuId"
    SKU_PRICE_ID = "SkuPriceId"

    LIST_UNIT_PRICE = "ListUnitPrice"
    LIST_COST = "ListCost"

    USAGE_QUANTITY = "UsageQuantity"
    USAGE_UNIT = "UsageUnit"


FOCUS_DATETIME_ISO_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


def get_dtype_for_focus_column_name(focus_column_name: FocusColumnNames):
    """
    Return the dtype for the focus column name
    """

    # convert loop to match statements
    if focus_column_name == FocusColumnNames.PLACE_HOLDER:
        raise ValueError("PLACE_HOLDER is not a valid focus column name")
    elif (
        focus_column_name == FocusColumnNames.CHARGE_PERIOD_START
        or focus_column_name == FocusColumnNames.CHARGE_PERIOD_END
        or focus_column_name == FocusColumnNames.BILLING_PERIOD_START
        or focus_column_name == FocusColumnNames.BILLING_PERIOD_END
    ):
        return pl.Datetime
    elif (
        focus_column_name == FocusColumnNames.BILLED_COST
        or focus_column_name == FocusColumnNames.EFFECTIVE_COST
        or focus_column_name == FocusColumnNames.LIST_COST
        or focus_column_name == FocusColumnNames.LIST_UNIT_PRICE
        or focus_column_name == FocusColumnNames.PRICING_QUANTITY
        or focus_column_name == FocusColumnNames.USAGE_QUANTITY
    ):
        return pl.Float64
    else:
        return pl.Utf8



import polars as pl

from focus_converter.configs.base_config import (
    ConversionPlan,
    StaticValueConversionArgs,
    UnnestValueConversionArgs,
    ValueMapConversionArgs,
)
from focus_converter.conversion_functions.validations import (
    STATIC_VALUE_COLUMN,
    ColumnValidator,
)
from focus_converter.models.focus_column_names import FocusColumnNames


class ColumnFunctions:
    @staticmethod
    def rename_column_functions(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ) -> pl.col:
        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(plan=plan, column_alias=column_alias)

        return pl.col(plan.column).alias(column_alias)

    @staticmethod
    def add_provider(provider, column_validator: ColumnValidator) -> pl.col:
        return pl.lit(provider).alias(FocusColumnNames.PROVIDER.value)

    @staticmethod
    def unnest(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ) -> pl.col:
        # validate conversion args
        if plan.conversion_args:
            conversion_args = UnnestValueConversionArgs.model_validate(
                plan.conversion_args
            )
        else:
            conversion_args = UnnestValueConversionArgs()

        # split column name by dots to find nested structs
        field_depths = plan.column.split(".")

        if conversion_args.children_type == "struct":
            predicate = pl.col(field_depths[0])
            for children in field_depths[1:]:
                predicate = predicate.struct.field(children)
        elif conversion_args.children_type == "list":
            # base predicate acts on each list element and collect child value
            # required base operation like min,max,sum etc can be applied on this base predicate.
            predicate = pl.col(field_depths[0]).list.eval(
                pl.element().struct.field(field_depths[1])
            )

            if conversion_args.aggregation_operation == "first":
                predicate = predicate.list.first()
            elif conversion_args.aggregation_operation == "last":
                predicate = predicate.list.last()
            elif conversion_args.aggregation_operation == "sum":
                predicate = predicate.list.sum()
            elif conversion_args.aggregation_operation == "mean":
                predicate = predicate.list.mean()
            elif conversion_args.aggregation_operation == "min":
                predicate = predicate.list.min()
            elif conversion_args.aggregation_operation == "max":
                predicate = predicate.list.max()
            else:
                raise RuntimeError(
                    f"Unknown aggregation_operation type: {conversion_args.aggregation_operation}"
                )
        else:
            raise RuntimeError(
                "Unknown children type: {}".format(conversion_args.children_type)
            )

        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(
            plan=plan, column_alias=column_alias, source_column=field_depths[0]
        )

        return predicate.alias(column_alias)

    @staticmethod
    def map_values(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ):
        # Converts values for a given dimension using a map. This map must also provide a default value.

        conversion_args = ValueMapConversionArgs.model_validate(plan.conversion_args)

        map_dict = {}
        for value_obj in conversion_args.value_list:
            map_dict.update({str(value_obj.key): value_obj.value})

        # if flag set, allow null value to be mapped to default value
        if conversion_args.apply_default_if_null:
            map_dict.update({None: conversion_args.default_value})
        else:
            map_dict.update({None: None})

        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(plan=plan, column_alias=column_alias)

        return (
            pl.col(plan.column)
            .replace(map_dict, default=conversion_args.default_value)
            .alias(column_alias)
        )

    @staticmethod
    def assign_static_value(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ) -> pl.col:
        conversion_args: StaticValueConversionArgs = (
            StaticValueConversionArgs.model_validate(plan.conversion_args)
        )

        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(
            plan=plan, column_alias=column_alias, source_column=STATIC_VALUE_COLUMN
        )

        return pl.lit(conversion_args.static_value).alias(column_alias)



import cProfile
import csv
import functools
import io
import pstats


class Profiler:
    def __init__(self, csv_format=False):
        self.csv_format = csv_format

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Wrap and execute profile
            profiler = cProfile.Profile()
            profiler.enable()
            result = func(*args, **kwargs)
            profiler.disable()
            profiling_result = pstats.Stats(profiler)

            generate_csv_file(args, profiling_result)

            generate_console_output(profiler)

            return result

        def generate_console_output(profiler):
            s = io.StringIO()
            sortby = "cumulative"
            ps = pstats.Stats(profiler, stream=s).sort_stats(sortby)
            ps.print_stats()
            print(s.getvalue())

        def generate_csv_file(args, profiling_result):
            if self.csv_format:
                # Determine the filename based on class and method name
                csv_filename = generate_file_name(args)
                with open(csv_filename, "w", newline="") as f:
                    w = csv.writer(f)
                    # Write the headers
                    headers = [
                        "ncalls",
                        "tottime",
                        "percall",
                        "cumtime",
                        "percall",
                        "filename:lineno(function)",
                    ]
                    w.writerow(headers)

                    # Write each row
                    for row in profiling_result.stats.items():
                        func_name, (cc, nc, tt, ct, callers) = row
                        w.writerow([nc, tt, tt / nc, ct, ct / cc, func_name])

        def generate_file_name(args):
            class_name = args[0].__class__.__name__ if args else "global"
            method_name = func.__name__
            base_filename = f"{class_name}_{method_name}_profile"
            csv_filename = f"{base_filename}.csv"
            return csv_filename

        return wrapper



from abc import ABC, abstractmethod

from focus_converter.conversion_functions.column_functions import ColumnFunctions
from focus_converter.conversion_functions.datetime_functions import (
    DateTimeConversionFunctions,
)
from focus_converter.conversion_functions.lookup_function import LookupFunction
from focus_converter.conversion_functions.sql_functions import SQLFunctions
from focus_converter.conversion_functions.string_functions import StringFunctions


# Define a Command interface with a method called execute().
class Command(ABC):
    @abstractmethod
    def execute(self):
        pass

    @abstractmethod
    def categorty(self):
        pass


# Date/Time Based Commands
class ConvertTimezoneCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            DateTimeConversionFunctions.convert_timezone(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "datetime"


class DateTimeConversionCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            DateTimeConversionFunctions.assign_timezone(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "datetime"


class DateTimeAssignUTCCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            DateTimeConversionFunctions.assign_utc_timezone(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "datetime"


class DateTimeMonthStartCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            DateTimeConversionFunctions.month_start(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "datetime"


class DateTimeMonthEndCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            DateTimeConversionFunctions.month_end(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "datetime"


class DateTimeParseDateTimeCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            DateTimeConversionFunctions.parse_datetime(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "datetime"


# Column Based Commands


class ColumnRenameCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            ColumnFunctions.rename_column_functions(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "column"


class ColumnUnnestCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            ColumnFunctions.unnest(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "column"


class ColumnMapValuesCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            ColumnFunctions.map_values(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "column"


class ColumnAssignStaticCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            ColumnFunctions.assign_static_value(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "column"


# SQL Based Commands


class SQLEvalQueryCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            SQLFunctions.eval_sql_query(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "sql"


class SQLEvalConditionsCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            SQLFunctions.eval_sql_conditions(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "sql"


# Lookup Based Commands
class LookupMapValuesCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        return column_exprs.append(
            LookupFunction.map_values_using_lookup(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "lookup"


# Deferred Column Based Commands
class DeferredColumnPlanApplyDefaultColumnCommand(Command):
    def execute(self, plan, column_alias, column_validator, deffered_column_function):
        return deffered_column_function.map_missing_column_plan(
            plan=plan,
            column_alias=column_alias,
            column_validator=column_validator,
        )

    def categorty(self):
        return "deferred"


class DeferredColumnMapDTypePlanCommand(Command):
    def execute(self, plan, column_alias, column_validator, deffered_column_function):
        return deffered_column_function.map_dtype_plan(
            plan=plan,
            column_validator=column_validator,
        )

    def categorty(self):
        return "deferred"


class StringFunctionsCommand(Command):
    def execute(self, plan, column_alias, column_validator, column_exprs):
        column_exprs.append(
            StringFunctions.convert(
                plan=plan,
                column_alias=column_alias,
                column_validator=column_validator,
            )
        )

    def categorty(self):
        return "string"



import multiprocessing
from typing import List

import polars as pl
import pyarrow.parquet as pq


def __writer_process__(export_path, queue: multiprocessing.Queue):
    while True:
        df = queue.get()
        if not isinstance(df, pl.DataFrame):
            break

        pq.write_to_dataset(
            root_path=export_path, compression="snappy", table=df.to_arrow()
        )


class DataExporter:
    def __init__(
        self,
        export_path,
        export_include_source_columns: bool,
        process_count: int = multiprocessing.cpu_count(),
    ):
        self.__export_path__ = export_path
        self.__export_include_source_columns__ = export_include_source_columns

        self.__queue__ = queue = multiprocessing.Queue(maxsize=process_count)

        self.__processes__ = processes = []
        for _ in range(process_count):
            p = multiprocessing.Process(
                target=__writer_process__,
                kwargs={"queue": queue, "export_path": self.__export_path__},
            )
            processes.append(p)

        # start processes
        [p.start() for p in processes]

    def __del__(self):
        if self.__queue__:
            self.close()

    def close(self):
        for _ in range(len(self.__processes__)):
            self.__queue__.put(None)

        for p in self.__processes__:
            p.join()
            p.close()
        del self.__queue__
        self.__queue__ = None

    def collect(self, lf: pl.LazyFrame, collected_columns: List[str]):
        if not self.__export_include_source_columns__:
            # collect only applied columns
            lf = lf.select(collected_columns)

        # compute final dataframe
        df: pl.DataFrame = lf.collect(streaming=True)
        self.__queue__.put(df)



import os
from typing import Annotated

import matplotlib.pyplot as plt
import numpy as np
import polars as pl
import typer

app = typer.Typer(
    name="FOCUS progress pie chart generator per providers", add_completion=False
)


def autopct_format(values):
    def my_format(pct):
        total = sum(values)
        val = int(round(pct * total / 100.0))
        return "{:.1f}%\n({v:d})".format(pct, v=val)

    return my_format


@app.command("generate", help="Generates progress pie chart for a providers")
def generate_for_all_providers(
    output_dir: Annotated[
        str,
        typer.Option(help="Write provider csvs to this path."),
    ],
    csv_rules_export_dir: Annotated[
        str,
        typer.Option(
            help="Provider csvs rules export to generate progress pie charts."
        ),
    ],
):
    for i, provider_csv_rule_export_name in enumerate(os.listdir(csv_rules_export_dir)):
        provider = provider_csv_rule_export_name.split(".")[0]
        csv_path = os.path.join(csv_rules_export_dir, provider_csv_rule_export_name)

        rules = pl.read_csv(csv_path)
        rules = rules.group_by(["FOCUS Dimension"]).agg(
            pl.col("Transform Type").first()
        )

        total_rule_count = rules.shape[0]

        pending = rules.filter(pl.col("Transform Type").eq("Not Defined"))

        pending_count = pending.shape[0]
        done_count = total_rule_count - pending_count

        y = np.array([pending_count, done_count])
        labels = ["Pending", "Done"]

        fig, ax = plt.subplots()
        ax.pie(y, labels=labels, autopct=autopct_format([pending_count, done_count]))
        ax.legend()
        ax.set_title(f"{provider.upper()} Progress")
        fig.savefig(f"{output_dir}/{provider}_progress_pie_chart.png")


if __name__ == "__main__":
    app()



import io
import json
import os

import typer
from focus_validator.validator import Validator
from PIL import Image
from rich import print
from typing_extensions import Annotated

from focus_converter.common.cli_options import (
    DATA_FORMAT_OPTION,
    DATA_PATH,
    EXPORT_INCLUDE_SOURCE_COLUMNS,
    EXPORT_PATH_OPTION,
    PARQUET_DATA_FORMAT_OPTION,
    PLAN_GRAPH_PATH,
    PROVIDER_OPTION,
)
from focus_converter.converter import FocusConverter
from focus_converter.data_loaders.data_loader import DataFormats

app = typer.Typer(name="FOCUS converters", add_completion=False)


@app.command("convert", help="Converts source cost data to FOCUS format")
def main(
    provider: PROVIDER_OPTION,
    export_path: EXPORT_PATH_OPTION,
    data_format: DATA_FORMAT_OPTION,
    data_path: DATA_PATH,
    parquet_data_format: PARQUET_DATA_FORMAT_OPTION = None,
    export_include_source_columns: EXPORT_INCLUDE_SOURCE_COLUMNS = True,
    column_prefix: Annotated[
        str,
        typer.Option(
            help="Optional prefix to add to generated column names",
            rich_help_panel="Column Prefix",
        ),
    ] = (None,),
    converted_column_prefix: Annotated[
        str,
        typer.Option(
            help="Optional prefix to add to generated column names",
            rich_help_panel="Column Prefix",
        ),
    ] = ((None,),),
    validate: Annotated[
        bool,
        typer.Option(
            help="Validate generated data to match FOCUS spec.",
            rich_help_panel="Validation",
        ),
    ] = False,
):
    # compute function for conversion

    if data_format == DataFormats.PARQUET and parquet_data_format is None:
        raise typer.BadParameter("parquet_data_format required")

    converter = FocusConverter(
        column_prefix=column_prefix, converted_column_prefix=converted_column_prefix
    )
    converter.load_provider_conversion_configs()
    converter.load_data(
        data_path=data_path,
        data_format=data_format,
        parquet_data_format=parquet_data_format,
    )
    converter.configure_data_export(
        export_path=export_path,
        export_include_source_columns=export_include_source_columns,
    )
    converter.prepare_horizontal_conversion_plan(provider=provider)
    converter.convert()

    if validate:
        for segment_file_name in os.listdir(export_path):
            file_path = os.path.join(export_path, segment_file_name)
            validator = Validator(
                data_filename=file_path,
                output_type="console",
                output_destination=None,
            )
            validator.load()
            validator.validate()
            break


@app.command("explain", help="Show conversion plan and saves a graph as an image")
def explain(provider: PROVIDER_OPTION, image_path: PLAN_GRAPH_PATH):
    # function to show conversion plan
    converter = FocusConverter()
    converter.load_provider_conversion_configs()
    converter.prepare_horizontal_conversion_plan(provider=provider)

    image = Image.open(io.BytesIO(converter.explain()))
    image.save(image_path)


@app.command("list-providers", help="List available providers")
def list_providers():
    converter = FocusConverter()
    converter.load_provider_conversion_configs()
    print(json.dumps({"providers": list(converter.plans.keys())}, indent=4))


if __name__ == "__main__":
    app()



import importlib.resources
import logging
import os
from operator import attrgetter
from typing import Dict, List, Optional

import polars as pl

from focus_converter.configs.base_config import ConversionPlan
from focus_converter.conversion_functions import STATIC_CONVERSION_TYPES
from focus_converter.conversion_functions.deferred_column_functions import (
    DeferredColumnFunctions,
)
from focus_converter.conversion_functions.sql_functions import SQLFunctions
from focus_converter.conversion_functions.validations import ColumnValidator
from focus_converter.conversion_strategy import (
    ColumnAssignStaticCommand,
    ColumnMapValuesCommand,
    ColumnRenameCommand,
    ColumnUnnestCommand,
    ConvertTimezoneCommand,
    DateTimeAssignUTCCommand,
    DateTimeConversionCommand,
    DateTimeMonthEndCommand,
    DateTimeMonthStartCommand,
    DateTimeParseDateTimeCommand,
    DeferredColumnMapDTypePlanCommand,
    DeferredColumnPlanApplyDefaultColumnCommand,
    LookupMapValuesCommand,
    SQLEvalConditionsCommand,
    SQLEvalQueryCommand,
    StringFunctionsCommand,
)
from focus_converter.data_loaders.data_exporter import DataExporter
from focus_converter.data_loaders.data_loader import DataLoader
from focus_converter.models.focus_column_names import (
    FocusColumnNames,
    get_dtype_for_focus_column_name,
)

# TODO: Make this path configurable so that we can load from a directory outside of the project
BASE_CONVERSION_CONFIGS = (
    importlib.resources.files("focus_converter")
    .joinpath("conversion_configs")
    .as_posix()
)


class FocusConverter:
    plans: Dict[str, List[ConversionPlan]]
    data_loader: DataLoader
    data_exporter: DataExporter = None

    # set of plan variables for horizontal transformation plans
    h_collected_columns: List[str]  # collected columns
    h_column_exprs: List[pl.col]  # column expressions
    h_sql_queries: List[str]  # sql_queries

    # temporary columns to be removed from final dataset
    __temporary_columns__: List[str]

    # column prefix used in source dataset
    __column_prefix__: Optional[str] = None

    # converted column prefix to be added to converted columns
    __converted_column_prefix__: Optional[str] = None

    def __init__(self, column_prefix=None, converted_column_prefix=None):
        self.__temporary_columns__ = []
        self.__column_prefix__ = column_prefix
        self.__converted_column_prefix__ = converted_column_prefix

        # ColumnValidator, used to validate column names in sql queries and transformations
        self.__column_validator__ = ColumnValidator()
        self.plans = {}

        # deferred column plans, these plans are applied after lazyframe is loaded
        self.__deferred_column_plans__ = DeferredColumnFunctions()

    def load_provider_conversion_configs(self):
        plans = {}

        for provider in os.listdir(BASE_CONVERSION_CONFIGS):
            # collect all plans specific to vendor
            provider_plans: List[ConversionPlan] = []

            provider_base_path = os.path.join(BASE_CONVERSION_CONFIGS, provider)
            for provider_config_name in os.listdir(provider_base_path):
                if not provider_config_name.endswith(".yaml"):
                    # ignores non yaml files, which may be reference datasets
                    continue

                provider_config_path = os.path.join(
                    provider_base_path, provider_config_name
                )
                logging.debug(f"Reading plan from path: {provider_config_name}")
                provider_plans.append(ConversionPlan.load_yaml(provider_config_path))

            plans[provider] = sorted(
                provider_plans,
                key=attrgetter("dimension_id", "priority"),
                reverse=False,
            )
        self.plans = plans

    def load_data(self, *args, **kwargs):
        self.data_loader = DataLoader(*args, **kwargs)

    def configure_data_export(self, *args, **kwargs):
        self.data_exporter = DataExporter(*args, **kwargs)

    def prepare_horizontal_conversion_plan(self, provider):
        # final set of columns produced after this transform step
        self.h_collected_columns = collected_columns = []

        # column expressions that will be applied to loaded lazy frame in order
        self.h_column_exprs = column_exprs = []

        # sql queries collected to be applied on the lazy frame
        self.h_sql_queries = sql_queries = []

        # lookup lazyframes arguments to be assembled later on the final source lazyframe
        self.lookup_reference_args = []

        # Create a dictionary to map conversion types to command classes.
        command_classes = {
            # Column based commands
            STATIC_CONVERSION_TYPES.CONVERT_TIMEZONE: ConvertTimezoneCommand,
            STATIC_CONVERSION_TYPES.ASSIGN_TIMEZONE: DateTimeConversionCommand,
            STATIC_CONVERSION_TYPES.ASSIGN_UTC_TIMEZONE: DateTimeAssignUTCCommand,
            STATIC_CONVERSION_TYPES.MONTH_START: DateTimeMonthStartCommand,
            STATIC_CONVERSION_TYPES.MONTH_END: DateTimeMonthEndCommand,
            STATIC_CONVERSION_TYPES.RENAME_COLUMN: ColumnRenameCommand,
            STATIC_CONVERSION_TYPES.MAP_VALUES: ColumnMapValuesCommand,
            STATIC_CONVERSION_TYPES.ASSIGN_STATIC_VALUE: ColumnAssignStaticCommand,
            # SQL based commands
            STATIC_CONVERSION_TYPES.SQL_QUERY: SQLEvalQueryCommand,
            STATIC_CONVERSION_TYPES.SQL_CONDITION: SQLEvalConditionsCommand,
            STATIC_CONVERSION_TYPES.PARSE_DATETIME: DateTimeParseDateTimeCommand,
            STATIC_CONVERSION_TYPES.UNNEST_COLUMN: ColumnUnnestCommand,
            # Lookup based commands
            STATIC_CONVERSION_TYPES.LOOKUP: LookupMapValuesCommand,
            # Deferred column plans
            STATIC_CONVERSION_TYPES.APPLY_DEFAULT_IF_COLUMN_MISSING: DeferredColumnPlanApplyDefaultColumnCommand,
            STATIC_CONVERSION_TYPES.SET_COLUMN_DTYPES: DeferredColumnMapDTypePlanCommand,
            # String based plans
            STATIC_CONVERSION_TYPES.STRING_FUNCTIONS: StringFunctionsCommand,
        }

        for plan in self.plans[provider]:
            # column name generated with temporary prefix
            column_alias = self.generate_column_name(plan)

            # add column to plan to collect these dimensions to be added in the computed dataframe
            if plan.focus_column != FocusColumnNames.PLACE_HOLDER:
                collected_columns.append(plan.focus_column.value)

            # process data based on conversion type
            command_class = command_classes.get(plan.conversion_type)
            if (
                command_class.categorty(self) == "column"
                or command_class.categorty(self) == "datetime"
            ):
                command_class().execute(
                    plan, column_alias, self.__column_validator__, column_exprs
                )
            elif command_class.categorty(self) == "sql":
                command_class().execute(
                    plan, column_alias, self.__column_validator__, sql_queries
                )
            elif command_class.categorty(self) == "lookup":
                command_class().execute(
                    plan,
                    column_alias,
                    self.__column_validator__,
                    self.lookup_reference_args,
                )
            elif command_class.categorty(self) == "deferred":
                command_class().execute(
                    plan,
                    column_alias,
                    self.__column_validator__,
                    self.__deferred_column_plans__,
                )
            elif command_class.categorty(self) == "string":
                command_class().execute(
                    plan,
                    column_alias,
                    self.__column_validator__,
                    column_exprs,
                )
            else:
                raise NotImplementedError(
                    f"Plan: {plan.conversion_type} not implemented"
                )

        # apply the plan to the lazy frame
        self.__column_validator__.validate_graph_is_connected()

        return column_exprs

    def generate_column_name(self, plan):
        if plan.column_prefix:
            column_alias = f"{plan.column_prefix}_{plan.focus_column.value}"
            self.__temporary_columns__.append(column_alias)
        else:
            column_alias = plan.focus_column.value
        return column_alias

    @staticmethod
    def __apply_sql_queries__(lf: pl.LazyFrame, sql_queries):
        for sql_query in sql_queries:
            sql_context = SQLFunctions.create_sql_context(lf=lf)
            lf = sql_context.execute(sql_query, eager=False)
        return lf

    @staticmethod
    def __apply_column_expressions__(
        lf: pl.LazyFrame, column_expressions: List[pl.col]
    ):
        for expr in column_expressions:
            lf = lf.with_columns_seq(expr)
        return lf

    @staticmethod
    def __apply_lookup_reference_plans__(lf: pl.LazyFrame, lookup_args):
        for lookup_arg in lookup_args:
            lf = lf.join(**lookup_arg)
        return lf

    def explain(self):
        # get batched data lazy frame, build the plan and then break
        return self.__column_validator__.generate_uml_graph()

    def apply_plan(self, lf: pl.LazyFrame) -> pl.LazyFrame:
        # creates lazy frame using the config, actual computation happens in collect

        lf = self.__apply_column_expressions__(
            lf=lf, column_expressions=self.h_column_exprs
        )

        # apply lazyframe joins
        lf = self.__apply_lookup_reference_plans__(
            lf=lf, lookup_args=self.lookup_reference_args
        )

        lf = self.__apply_sql_queries__(lf=lf, sql_queries=self.h_sql_queries)

        # drop temporary columns from the final dataset
        lf = lf.drop(self.__temporary_columns__)

        return lf

    def __re_map_source_columns__(self, lf: pl.LazyFrame):
        # helper function to re-map prefixed source columns to source columns
        # so that conversion plans can be applied.

        column_prefix_length = len(self.__column_prefix__)
        for column in lf.columns:
            if column.startswith(self.__column_prefix__):
                orig_column_name = column[column_prefix_length:]
                lf = lf.with_columns_seq(pl.col(column).alias(orig_column_name))

                # add remapped column to temporary columns list so that it can be dropped later
                self.__temporary_columns__.append(orig_column_name)
        return lf

    def __add_empty_columns_for_missing_focus_columns__(self, lf: pl.LazyFrame):
        # add missing focus columns with null values to produce a valid parquet file
        for focus_column_name in FocusColumnNames:
            if (
                focus_column_name.value not in self.h_collected_columns
                and focus_column_name != FocusColumnNames.PLACE_HOLDER
            ):
                lf = lf.with_columns(
                    pl.lit(None)
                    .alias(focus_column_name.value)
                    .cast(get_dtype_for_focus_column_name(focus_column_name))
                )
        return lf

    def __process_lazy_frame__(self, lf: pl.LazyFrame):
        # prepares lazyframe for the operations to be applied on the lazy loaded polars dataframe
        if self.__column_prefix__ is not None:
            lf = self.__re_map_source_columns__(lf=lf)

        # apply deferred column plans
        lf = self.__deferred_column_plans__.apply_missing_column_plan(lf=lf)
        lf = self.__deferred_column_plans__.apply_dtype_plan(lf=lf)

        # validate all source columns exist in the lazy frame
        self.__column_validator__.validate_lazy_frame_columns(lf=lf)

        # add missing focus columns with null values to produce a valid parquet file
        lf = self.__add_empty_columns_for_missing_focus_columns__(lf=lf)

        return self.apply_plan(lf=lf)

    def convert(self):
        error = None

        for lf in self.data_loader.data_scanner():
            try:
                lf = self.__process_lazy_frame__(lf=lf)
                self.data_exporter.collect(
                    lf=lf, collected_columns=list(set(self.h_collected_columns))
                )
            except Exception as e:
                error = e
                break

        if error is not None:
            # cleanup data export now that we have to close all subprocesses
            if self.data_exporter:
                self.data_exporter.close()

            raise error



from enum import Enum


class STATIC_CONVERSION_TYPES(Enum):
    # datetime functions
    CONVERT_TIMEZONE = "convert_timezone"
    ASSIGN_TIMEZONE = "assign_timezone"
    ASSIGN_UTC_TIMEZONE = "assign_utc_timezone"
    PARSE_DATETIME = "parse_datetime"
    MONTH_START = "month_start"
    MONTH_END = "month_end"

    # sql rule functions
    SQL_QUERY = "sql_query"
    SQL_CONDITION = "sql_condition"

    # column rename function
    RENAME_COLUMN = "rename_column"

    # unnest operation
    UNNEST_COLUMN = "unnest"

    # lookup operation
    LOOKUP = "lookup"

    # value mapping function
    MAP_VALUES = "map_values"

    # allows setting static values
    ASSIGN_STATIC_VALUE = "static_value"

    # apply default values if column not present
    APPLY_DEFAULT_IF_COLUMN_MISSING = "apply_default_if_column_missing"

    # set column dtypes
    SET_COLUMN_DTYPES = "set_column_dtypes"

    # string functions
    STRING_FUNCTIONS = "string_functions"


__all__ = [
    "STATIC_CONVERSION_TYPES",
]



from typing import List, Literal

from pydantic import BaseModel


class StringTransformArgs(BaseModel):
    steps: List[
        Literal[
            "lower",
            "upper",
            "title",
        ]
    ]
